#include "physics_entity.h"
#include "camera.h"
#include "rigidbody.h"
#include "shape.h"
#include "render.h"
#include "debug.h"
#include "program.h"

void physics_entity_debug_draw(PhysicsEntity self) {
    RigidBody* body = self.tc->get_rigidbody(self.data);
    Shape* shape = self.tc->get_shape(self.data);
    Transform* transform = self.transformable->get_transform(self.data);

    shape_draw(shape, *transform);
    rigidbody_debug_draw_contacts(body);

    Vector lhs = transform->position;
    Vector rhs = vaddf(lhs, rigidbody_get_velocity(body));
    lhs = camera_world_to_pixel_point(&g_camera, lhs);
    rhs = camera_world_to_pixel_point(&g_camera, rhs);
    SDL_SetRenderDrawColor(g_renderer, 0, 255, 0, 255);
    SDL_RenderDrawLine(g_renderer, lhs.x, lhs.y, rhs.x, rhs.y);

    rhs = camera_world_to_pixel_point(&g_camera, vaddf(transform->position, rigidbody_get_force(body)));
    SDL_SetRenderDrawColor(g_renderer, 0, 255, 255, 255);
    SDL_RenderDrawLine(g_renderer, lhs.x, lhs.y, rhs.x, rhs.y);
}
static  inline
Vector _internal_calculate_contact_force(RigidBody* self, Contact* contact) {
    Collision hit = contact->hit;

    const Vector velocity = hit.velocity;
    const Vector normal = hit.normal;

    const float elasticity = 4.0;
    // const float damping = 1.0;//4.f;

    return vmulff(normal, elasticity * -vdotf(normal, velocity));
}

static inline
void _internal_default_contact_solver(RigidBody* body, Contact* contact, Vector* solve_forces) {
    Collision hit = contact->hit;
    if(vdotf(hit.normal, rigidbody_get_velocity(body)) > 0.0)
        return;

    float normalmag = vmagnitudef(contact->hit.normal);
    ASSERT_RETURN(fabsf(normalmag - 1.f) < 0.99999f,, "Normal of collision not one");

    Vector force = _internal_calculate_contact_force(body, contact);
    float dot = vdotf(vnormalizedf(*solve_forces), force);
    if (veqf(*solve_forces, ZeroVector) || dot * dot > vsqrmagnitudef(*solve_forces) || dot <= 0.0f) {
        *solve_forces = vaddf(*solve_forces, force);
        rigidbody_add_impulse(body, force, 1);
    }

    ASSERT_RETURN(!visnanf(force),, "Force contains NaN (1)");
}

void default_contact_solver(PhysicsEntity self, List* contacts) {
    Vector solve_forces = ZeroVector;
    list_foreach(Contact, contact, contacts) {
        _internal_default_contact_solver(self.tc->get_rigidbody(self.data), contact, &solve_forces);
    }
}

void physics_entity_handle_contacts(PhysicsEntity self) {
    RigidBody* body = self.tc->get_rigidbody(self.data);

    ASSERT_RETURN(!visnanf(rigidbody_get_velocity(body)),, "Velocity is NaN (0)");

    rigidbody_collect_contacts(body);
    if(rigidbody_get_contacts(body)->len > 0)
        self.tc->collision_solver(self.data, rigidbody_get_contacts(body));

    ASSERT_RETURN(!visnanf(rigidbody_get_velocity(body)),, "Velocity is NaN (1)");
}
